#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include "types.h"

#if defined(LITTLE_ENDIAN)
#define PACK_COLOR24(r, g, b)	((r << 16) | (g << 8) | b)
#else
#define PACK_COLOR24(r, g, b)	(r | (g << 8) | (b << 16))
#endif

static void *load_ppm(FILE *fp, unsigned long *xsz, unsigned long *ysz);

int check_ppm(FILE *fp) {
	fseek(fp, 0, SEEK_SET);
	if(fgetc(fp) == 'P' && fgetc(fp) == '6') {
		return 1;
	}
	return 0;
}

static int read_to_wspace(FILE *fp, char *buf, int bsize) {
	int c, count = 0;
	
	while((c = fgetc(fp)) != -1 && !isspace(c) && count < bsize - 1) {
		if(c == '#') {
			while((c = fgetc(fp)) != -1 && c != '\n' && c != '\r');
			c = fgetc(fp);
			if(c == '\n' || c == '\r') continue;
		}
		*buf++ = c;
		count++;
	}
	*buf = 0;
	
	while((c = fgetc(fp)) != -1 && isspace(c));
	ungetc(c, fp);
	return count;
}

void *load_image(const char *fname, unsigned long *xsz, unsigned long *ysz) {
	FILE *fp;

	if(!(fp = fopen(fname, "rb"))) {
		fprintf(stderr, "failed to open %s: %s\n", fname, strerror(errno));
		return 0;
	}

	if(!check_ppm(fp)) {
		fprintf(stderr, "file %s is not a portable pixmap\n", fname);
		return 0;
	}

	return load_ppm(fp, xsz, ysz);
}

static void *load_ppm(FILE *fp, unsigned long *xsz, unsigned long *ysz) {
	char buf[64];
	int bytes, raw;
	unsigned int w, h, i, sz;
	uint32_t *pixels;
	
	fseek(fp, 0, SEEK_SET);
	
	bytes = read_to_wspace(fp, buf, 64);
	raw = buf[1] == '6';

	if((bytes = read_to_wspace(fp, buf, 64)) == 0) {
		fclose(fp);
		return 0;
	}
	if(!isdigit(*buf)) {
		fprintf(stderr, "load_ppm: invalid width: %s", buf);
		fclose(fp);
		return 0;
	}
	w = atoi(buf);

	if((bytes = read_to_wspace(fp, buf, 64)) == 0) {
		fclose(fp);
		return 0;
	}
	if(!isdigit(*buf)) {
		fprintf(stderr, "load_ppm: invalid height: %s", buf);
		fclose(fp);
		return 0;
	}
	h = atoi(buf);

	if((bytes = read_to_wspace(fp, buf, 64)) == 0) {
		fclose(fp);
		return 0;
	}
	if(!isdigit(*buf) || atoi(buf) != 255) {
		fprintf(stderr, "load_ppm: invalid or unsupported max value: %s", buf);
		fclose(fp);
		return 0;
	}

	if(!(pixels = malloc(w * h * sizeof *pixels))) {
		fprintf(stderr, "load_ppm: malloc failed");
		fclose(fp);
		return 0;
	}

	sz = h * w;
	for(i=0; i<sz; i++) {
		int r = fgetc(fp);
		int g = fgetc(fp);
		int b = fgetc(fp);

		if(r == -1 || g == -1 || b == -1) {
			free(pixels);
			fclose(fp);
			fprintf(stderr, "load_ppm: EOF while reading pixel data");
			return 0;
		}
		pixels[i] = PACK_COLOR24(r, g, b);
	}

	fclose(fp);

	if(xsz) *xsz = w;
	if(ysz) *ysz = h;
	return pixels;
}

#if 0
int save_ppm(FILE *fp, void *pixels, unsigned long xsz, unsigned long ysz) {
	int i, sz = xsz * ysz;
	uint32_t *ptr = pixels;

	fprintf(fp, "P6\n%lu %lu\n255\n", xsz, ysz);
	/*fprintf(fp, "P3\n%lu %lu\n255\n", xsz, ysz);*/

	for(i=0; i<sz; i++) {
		int r = (*ptr & RED_MASK32) >> RED_SHIFT32;
		int g = (*ptr & GREEN_MASK32) >> GREEN_SHIFT32;
		int b = (*ptr++ & BLUE_MASK32) >> BLUE_SHIFT32;

		fputc(r, fp);
		fputc(g, fp);
		if(fputc(b, fp) == EOF) {
			fprintf(stderr, "save_ppm: failed to write to file");
			fclose(fp);
			return -1;
		}

		/*fprintf(fp, "%d %d %d\n", r, g, b);*/
	}

	fclose(fp);
	return 0;
}
#endif
