#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <SDL.h>
#include "mini3d.h"

void init(void);
void redraw(unsigned int msec);
void slime(uint32_t *fb, uint32_t *img);
void clean_up(void);
unsigned int timer(void);

void draw_cube(float sz, float r, float g, float b, float spow);
void draw_plane(float xsz, float ysz, int subdiv, float r, float g, float b);

void *load_image(const char *fname, unsigned long *xsz, unsigned long *ysz);

#define ASPECT	(16.0 / 9.0)

int xres = 640;
int yres = 480;
int fullscreen;
int vp_x, vp_y, slime_vp_x, slime_vp_y, voffset;

int hack = 152;
int use_slime = 1;

unsigned int floor_tex;
SDL_Surface *fbsurf;

uint32_t *logo_pix;
unsigned long logo_x, logo_y;

#define SLIME_FRAMES	110
uint32_t *frames[SLIME_FRAMES];

#define FRAME_INTERVAL	27

int main(int argc, char **argv) {
	int i;
	unsigned int msec = 0, frame_count = 0, prev_time = 0;

	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-' && argv[i][2] == 0) {
			char *x;
			
			switch(argv[i][1]) {
			case 'f':
				fullscreen = 1;
				break;
				
			case 's':
				if(!isdigit(argv[++i][0]) || !(x = strchr(argv[i], 'x')) || !isdigit(*(x + 1))) {
					fprintf(stderr, "-s must be followed by WIDTHxHEIGHT\n");
					return EXIT_FAILURE;
				}
				xres = atoi(argv[i]);
				yres = atoi(x + 1);				
				break;

			default:
				fprintf(stderr, "invalid option: %s\n", argv[i]);
				return EXIT_FAILURE;
			}
		}
	}
	
	init();
	atexit(clean_up);

	while(1) {
		SDL_Event event;

		while(SDL_PollEvent(&event)) {
			switch(event.type) {
			case SDL_KEYDOWN:
				switch(event.key.keysym.sym) {
				case 27:
					if(msec > 1000) {
						printf("fps: %.3f (limiter was: %.3f)\n", (float)frame_count / (float)(msec / 1000), 1000.0 / FRAME_INTERVAL);
					}
					exit(0);

				case 'a':
					hack++;
					printf("%d\n", hack);
					break;

				case 'z':
					hack--;
					if(hack < 0) hack = 0;
					printf("%d\n", hack);
					break;

				case 's':
					use_slime = !use_slime;
					break;

				default:
					break;
				}
				break;

				
			default:
				break;
			}
		}

		msec = timer();
		if(msec - prev_time >= FRAME_INTERVAL) {
			redraw(msec);
			frame_count++;
			prev_time = msec;
		}
	}
}

void init(void) {
	int i;
	float amb[] = {0.3, 0.3, 0.3, 0.0};
	float lpos[] = {-100, 100, 100, 1.0};

	vp_x = xres;
	vp_y = (float)xres / ASPECT;
	slime_vp_x = vp_x / 2;
	slime_vp_y = vp_y;
	voffset = (yres - slime_vp_y) / 2 * xres;

	printf("allocating: %d mb for slime buffers\n", (int)(slime_vp_x * slime_vp_y * SLIME_FRAMES * sizeof(uint32_t)) / 1024 / 1024);

	for(i=0; i<SLIME_FRAMES; i++) {
		if(!(frames[i] = calloc(slime_vp_x * slime_vp_y, sizeof *frames[i]))) {
			perror("malloc failure");
			exit(EXIT_FAILURE);
		}
	}
	
	SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER);
	if(!(fbsurf = SDL_SetVideoMode(xres, yres, 32, SDL_SWSURFACE | (fullscreen ? SDL_FULLSCREEN : 0)))) {
		fprintf(stderr, "SDL init failed\n");
		exit(EXIT_FAILURE);
	}
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY / 2, SDL_DEFAULT_REPEAT_INTERVAL);
	SDL_WM_SetCaption("Slimy vectors - Nuclear / The Lab", 0);

	m3d_init();
	m3d_viewport(0, 0, vp_x, vp_y);

	m3d_clear_color(0, 0, 0, 0);
	m3d_clear_depth(1);

	m3d_disable(M3D_DEPTH_TEST);
	m3d_shade_model(M3D_FLAT);

	m3d_enable(M3D_LIGHTING);
	m3d_enable(M3D_LIGHT0);
	m3d_light(M3D_LIGHT0, M3D_POSITION, lpos);

	m3d_light_modelv(M3D_LIGHT_MODEL_AMBIENT, amb);

	m3d_matrix_mode(M3D_PROJECTION);
	m3d_load_identity();
	m3d_perspective(45.0, (float)vp_x / (float)vp_y, 1.0, 100.0);
	m3d_matrix_mode(M3D_MODELVIEW);

	{
		int x, y;
		uint32_t *floor_pixels;
		int tex_x = 64, tex_y = 64;

		if(!(floor_pixels = malloc(tex_x * tex_y * sizeof *floor_pixels))) {
			perror("malloc failed");
			exit(EXIT_FAILURE);
		}

		for(y=0; y<tex_y; y++) {
			for(x=0; x<tex_x; x++) {
				floor_pixels[y * tex_x + x] = ((x / 8 % 2) == (y / 8 % 2)) ? 0xffaa40 : 0x80ffa0;
			}
		}

		m3d_gen_textures(1, &floor_tex);
		m3d_bind_texture(M3D_TEXTURE_2D, floor_tex);
		m3d_tex_image(M3D_TEXTURE_2D, 0, 3, tex_x, tex_y, 0, M3D_BGRA, M3D_UNSIGNED_BYTE, floor_pixels);

		free(floor_pixels);
	}

	if(!(logo_pix = load_image("title.ppm", &logo_x, &logo_y))) {
		exit(EXIT_FAILURE);
	}

	SDL_FillRect(fbsurf, 0, 0x000060);
}

void clean_up(void) {
	m3d_destroy();
	SDL_Quit();
}

#define SZ		4.5
#define HSZ		(SZ / 2.0)
void redraw(unsigned int msec) {
	static int first_call = 1;
	
	int xoffs = (xres - slime_vp_x) / 2;
	float t = (float)msec / 9.0;

	if(SDL_MUSTLOCK(fbsurf)) SDL_LockSurface(fbsurf);

	/* --------- draw the background --------- */
	m3d_clear(M3D_COLOR_BUFFER_BIT);

	m3d_load_identity();
	m3d_translate(0, -1, -3.5);
	m3d_rotate(t / 4.0, 0, 1, 0);
	
	m3d_disable(M3D_LIGHTING);
	m3d_enable(M3D_TEXTURE_2D);
	draw_plane(SZ, SZ, 7, 0.8, 0.8, 0.8);
	m3d_disable(M3D_TEXTURE_2D);
	m3d_enable(M3D_LIGHTING);

	memcpy((uint32_t*)fbsurf->pixels + voffset, m3d_get_pixel_data(), vp_x * vp_y * sizeof(uint32_t));

	/* --------- draw the slimy bit ---------- */
	m3d_clear(M3D_COLOR_BUFFER_BIT);

	m3d_load_identity();
	m3d_translate(0, 0, -3.5);
	m3d_rotate(cos(t/150.0) * 45.0, 1, 0, 0);
	m3d_rotate(cos(t/200.0) * 90.0 + sin(t / 100), 0, 1, 0);

	draw_cube(1.0, 1, 0.6, 0.4, 20);

	slime((uint32_t*)fbsurf->pixels + voffset + xoffs, m3d_get_pixel_data() + xoffs);

	if(first_call) memcpy(fbsurf->pixels, logo_pix, logo_x * logo_y * sizeof(uint32_t));
	/* --------- done drawing --------- */
	if(SDL_MUSTLOCK(fbsurf)) SDL_UnlockSurface(fbsurf);

	SDL_Flip(fbsurf);
	first_call = 0;
}

#define HANDLE_COLORKEY
void slime(uint32_t *fb, uint32_t *img) {
	static int current;
	int i, j;

	for(i=0; i<slime_vp_y; i++) {
		memcpy(frames[current] + i * slime_vp_x, img, slime_vp_x * sizeof(uint32_t));
		img += vp_x;
	}

	for(i=0; i<slime_vp_y; i++) {
		int k = (i + hack) / 2;
#ifdef HANDLE_COLORKEY
		uint32_t *src = frames[(current + k) % SLIME_FRAMES] + i * slime_vp_x;

		for(j=0; j<slime_vp_x; j++) {
			if(*src) {
				fb[j] = *src;
			}
			src++;
		}
#else
		memcpy(fb, frames[(current + k) % SLIME_FRAMES] + i * slime_vp_x, slime_vp_x * sizeof(uint32_t));
#endif
		fb += xres;
	}

	current = (current + 1) % SLIME_FRAMES;
}

unsigned int timer(void) {
	static unsigned int start_time;
	if(!start_time) {
		start_time = SDL_GetTicks();
	}
	return SDL_GetTicks() - start_time;
}

void draw_cube(float sz, float r, float g, float b, float spow) {
	float color[4];
	float white[] = {1, 1, 1, 1};
	sz *= 0.5;

	color[0] = r;
	color[1] = g;
	color[2] = b;
	color[3] = 1.0;

	m3d_materialv(M3D_AMBIENT_AND_DIFFUSE, color);
	m3d_materialv(M3D_SPECULAR, white);
	m3d_material(M3D_SHININESS, spow);
	
	m3d_begin(M3D_QUADS);
	
	/* face +Z */
	m3d_normal(0, 0, 1);
	m3d_color3(1, 0, 0);
	m3d_tex_coord(0, 0);
	m3d_vertex(sz, -sz, sz);
	m3d_tex_coord(1, 0);
	m3d_vertex(-sz, -sz, sz);
	m3d_tex_coord(1, 1);
	m3d_vertex(-sz, sz, sz);
	m3d_tex_coord(0, 1);
	m3d_vertex(sz, sz, sz);

	/* face -X */
	m3d_normal(-1, 0, 0);
	m3d_color3(1, 0, 0);
	m3d_tex_coord(0, 0);
	m3d_vertex(-sz, -sz, sz);
	m3d_tex_coord(1, 0);
	m3d_vertex(-sz, -sz, -sz);
	m3d_tex_coord(1, 1);
	m3d_vertex(-sz, sz, -sz);
	m3d_tex_coord(0, 1);
	m3d_vertex(-sz, sz, sz);

	/* face -Z */
	m3d_normal(0, 0, -1);
	m3d_color3(1, 0, 0);
	m3d_tex_coord(0, 1);
	m3d_vertex(-sz, sz, -sz);
	m3d_tex_coord(0, 0);
	m3d_vertex(-sz, -sz, -sz);
	m3d_tex_coord(1, 0);
	m3d_vertex(sz, -sz, -sz);
	m3d_tex_coord(1, 1);
	m3d_vertex(sz, sz, -sz);

	/* face +X */
	m3d_normal(1, 0, 0);
	m3d_color3(1, 0, 0);
	m3d_tex_coord(0, 1);
	m3d_vertex(sz, sz, -sz);
	m3d_tex_coord(0, 0);
	m3d_vertex(sz, -sz, -sz);
	m3d_tex_coord(1, 0);
	m3d_vertex(sz, -sz, sz);
	m3d_tex_coord(1, 1);
	m3d_vertex(sz, sz, sz);

	/* face +Y */
	m3d_normal(0, 1, 0);
	m3d_color3(1, 0, 0);
	m3d_tex_coord(1, 0);
	m3d_vertex(sz, sz, sz);
	m3d_tex_coord(0, 0);
	m3d_vertex(-sz, sz, sz);
	m3d_tex_coord(0, 1);
	m3d_vertex(-sz, sz, -sz);
	m3d_tex_coord(1, 1);
	m3d_vertex(sz, sz, -sz);

	/* face -Y */
	m3d_normal(0, -1, 0);
	m3d_color3(1, 0, 0);
	m3d_tex_coord(1, 0);
	m3d_vertex(-sz, -sz, -sz);
	m3d_tex_coord(1, 1);
	m3d_vertex(-sz, -sz, sz);
	m3d_tex_coord(0, 1);
	m3d_vertex(sz, -sz, sz);
	m3d_tex_coord(0, 0);
	m3d_vertex(sz, -sz, -sz);

	m3d_end();
}

void draw_plane(float xsz, float ysz, int subdiv, float r, float g, float b) {
	int i, j;
	int quads_row = subdiv + 1;		/* quads per row */
	int vrow = subdiv + 2;		/* vertices per row */

	float du = 1.0 / (float)quads_row;
	float dv = 1.0 / (float)quads_row;

	float color[4];
	float black[] = {0, 0, 0, 0};

	color[0] = r;
	color[1] = g;
	color[2] = b;
	color[3] = 1;

	m3d_materialv(M3D_AMBIENT_AND_DIFFUSE, color);
	m3d_materialv(M3D_SPECULAR, black);

	m3d_begin(M3D_QUADS);
	m3d_color3(r, g, b);

	for(j=0; j<quads_row; j++) {
		for(i=0; i<quads_row; i++) {
			float x, y;
			float u = (float)i / (float)quads_row;
			float v = (float)j / (float)quads_row;

			x = (u - 0.5) * xsz;
			y = (1.0 - v - 0.5) * ysz;
			m3d_tex_coord(u, 1.0 - v);
			m3d_vertex(x, 0, y);
			
			x = (u - 0.5) * xsz;
			y = (1.0 - (v + dv) - 0.5) * ysz;
			m3d_tex_coord(u, 1.0 - (v + dv));
			m3d_vertex(x, 0, y);

			x = (u + du - 0.5) * xsz;
			y = (1.0 - (v + dv) - 0.5) * ysz;
			m3d_tex_coord(u + du, 1.0 - (v + dv));
			m3d_vertex(x, 0, y);

			x = (u + du - 0.5) * xsz;
			y = (1.0 - v - 0.5) * ysz;
			m3d_tex_coord(u + du, 1.0 - v);
			m3d_vertex(x, 0, y);
		}
	}

	m3d_end();
}

